#include <math.h>
#include "mex.h"

/**************************************************/
/* fonction implementant la methode de kchemin    */
/* mars 2006                                      */
/* F Suard                                        */
/**************************************************/
#define max(a,b)(a>b?a:b)
#define min(a,b)(a<b?a:b)

void calculerNoyauNoeuds(mxArray * , mxArray *, double *, double *);
void calculerNoyauArcs(mxArray * , mxArray *, double *, double *);
double comparerChemins(mxArray *, mxArray * , double *, double *, int, int);

/*double min(double a, double b)
{
    if(a < b)
        return a;
    else
        return b;
}

double max(double a, double b)
{
    if(a>b)
        return a;
    else
        return b;
}*/

void mexFunction( int nlhs, mxArray *plhs[],  int nrhs, const mxArray *prhs[])
{
    int i,j,k1,valeur,valeur2,valeur3, indice1, indice2, longueur;
    double *sigma;
    mxArray *etiquetteNoeud1, *etiquetteArc1, *etiquetteNoeud2, *etiquetteArc2, *chemin1, *chemin2, *MDist1, *MDist2;
    int nbEtiquettesNoeuds, nbEtiquettesArcs, nbNoeuds1, nbNoeuds2, nbArcs1, nbArcs2;
    double *sigmaNoeud, *sigmaArc, *noyauArc, *noyauNoeud;
    int lmax, lmax1,lmax2;
    double max1,max2,maxi1,maxi2;
    double *histoChemins1, *histoChemins2;
	double somme,temp,K=0.0;

  /*verification des entrees sorties */
  if(nlhs > 1)
  {
    mexErrMsgTxt("Incorrect number of output");
  }

  if(nrhs > 4)
  {
    mexErrMsgTxt("Incorrect number of input");
  }
    

   etiquetteNoeud1 = mxGetField(prhs[0],0,"etiquetteNoeud");
   etiquetteArc1 = mxGetField(prhs[0],0,"etiquetteArc");
   etiquetteNoeud2 = mxGetField(prhs[1],0,"etiquetteNoeud");
   etiquetteArc2 = mxGetField(prhs[1],0,"etiquetteArc");

  /*verification du nombre d'etiquettes des arcs et noeuds*/
  if(mxGetN(etiquetteNoeud1) != mxGetN(etiquetteNoeud2))
  {
    mexErrMsgTxt("Nombre d etiquettes dans les noeuds different");
  }

   if(mxGetN(etiquetteArc1) != mxGetN(etiquetteArc2))
  {
    mexErrMsgTxt("Nombre d etiquettes dans les arcs different");
  }

  nbEtiquettesNoeuds = mxGetN(etiquetteNoeud1);
  nbEtiquettesArcs = mxGetN(etiquetteArc1);
  nbNoeuds1 = mxGetM(etiquetteNoeud1);
  nbNoeuds2 = mxGetM(etiquetteNoeud2);
  nbArcs1 = mxGetM(etiquetteArc1);
  nbArcs2 = mxGetM(etiquetteArc2);

 
  sigmaNoeud = malloc(nbEtiquettesNoeuds*sizeof(double));
  sigmaArc = malloc(nbEtiquettesArcs*sizeof(double));

  /*verification du nombre de valeurs dans sigma. si une seule valeur, elle est dupliquee pour  les noeuds et les arcs */
  sigma = mxGetPr(prhs[2]);
  if(mxGetN(prhs[2]) == 1)
  {
    for(i=0;i<nbEtiquettesNoeuds;i++)
    {
        sigmaNoeud[i] = *(sigma);/* *(mxGetPr(prhs[2]));*/
    }
    for(i=0;i<nbEtiquettesArcs;i++)
    {
        sigmaArc[i] = *(sigma);/* *(mxGetPr(prhs[2]));*/
    }
  }
  else if(mxGetN(prhs[2]) == (nbEtiquettesNoeuds+nbEtiquettesArcs))
  {
    for(i=0;i<nbEtiquettesNoeuds;i++)
    {
        sigmaNoeud[i] = *(mxGetPr(prhs[2])+i);
    }
    for(i=0;i<nbEtiquettesArcs;i++)
    {
        sigmaArc[i] = *(mxGetPr(prhs[2])+i+nbEtiquettesNoeuds-1);
    } 
  }
  else 
  {
    mexErrMsgTxt("Taille de sigma incorrecte");
  }
  
  /* calcul du noyau de noeuds */
  noyauNoeud = malloc(nbNoeuds1*nbNoeuds2*sizeof(double));
  calculerNoyauNoeuds(etiquetteNoeud1 , etiquetteNoeud2, sigmaNoeud, noyauNoeud);
  
  /* calcul du noyau d'arcs */
  noyauArc = malloc(nbArcs1*nbArcs2*sizeof(double));
  calculerNoyauArcs(etiquetteArc1 , etiquetteArc2, sigmaArc, noyauArc);

  /*for(i=0;i<nbArcs1;i++)
    {
        for(j=0;j<nbArcs2;j++)
        {
           mexPrintf("%f ",*((noyauArc)+j*nbArcs1+i));
        }
        mexPrintf("\n");
    }*/
  
  
  
  /*taille du chemin le plus long*/
  chemin1 = mxGetField(prhs[0],0,"Chemins");
  chemin2 = mxGetField(prhs[1],0,"Chemins");
  MDist1 = mxGetField(prhs[0],0,"MDist");
  MDist2 = mxGetField(prhs[1],0,"MDist");
  lmax = 0;
  lmax1= 0;
  lmax2 = 0;

  for(i=0;i<mxGetN(MDist1)*mxGetM(MDist1);i++)
  {
        valeur= (int)*(mxGetPr(MDist1)+i);
        if( valeur > lmax1)
              lmax1 = valeur;        
  }
    

   histoChemins1 = malloc((lmax1+1)*sizeof(double));
    for(i=0;i<=lmax1;i++)
        histoChemins1[i] = 0;

	for(i=0;i<mxGetN(MDist1)*mxGetM(MDist1);i++)
     {     
		valeur= (int)*(mxGetPr(MDist1)+i);
		histoChemins1[valeur] = histoChemins1[valeur]+1;    
     }
		
	for(i=0;i<mxGetN(MDist2)*mxGetM(MDist2);i++)
	{
       valeur= (int)*(mxGetPr(MDist2)+i);
        if( valeur > lmax2)
              lmax2 = valeur;      
	}
    
    histoChemins2 = malloc((lmax2+1)*sizeof(double));
    for(i=0;i<=lmax2;i++)
        histoChemins2[i] = 0;

    for(i=0;i<mxGetN(MDist2)*mxGetM(MDist2);i++)
     {     
		valeur= (int)*(mxGetPr(MDist2)+i);
		histoChemins2[valeur] = histoChemins2[valeur]+1;    
     }
	
	/* longueur maximum des chemins*/	
	if(nrhs > 3)
    {
        lmax = *(mxGetPr(prhs[3]));
        lmax = min(lmax,lmax1); 
        lmax = min(lmax,lmax2);
         
    }
    else
    {
        lmax = min(lmax1,lmax2);
    }
    
    for(longueur=0;longueur<=lmax;longueur++)
	{
	if(histoChemins1[longueur]>0 && histoChemins2[longueur]> 0)
	{
		somme=0;
        maxi1=-100000000;
        maxi2=-100000000;
        for(i=0;i<nbNoeuds1*nbNoeuds1;i++) /*nbNoeuds1*nbNoeuds1*/
        {
            valeur = *(mxGetPr(MDist1)+i);
            if(valeur == longueur)
            {
                for(k1=0;k1<nbNoeuds2*nbNoeuds2;k1++) 
                {
                    valeur3 = *(mxGetPr(MDist2)+k1);
                    if( valeur3 == longueur)
                    {
                        temp = comparerChemins(mxGetCell(chemin1,i), mxGetCell(chemin2,k1), noyauNoeud, noyauArc,nbNoeuds1,nbArcs1);
                        if( maxi1 < temp)
                        {
                            maxi1=temp;
                        }
                    }
                }
                    somme=somme+maxi1;
                maxi1=-100000000;
            }
        }
       for(j=0;j<nbNoeuds2*nbNoeuds2;j++) 
        {
            valeur2 = *(mxGetPr(MDist2)+j);
            if( valeur2 == longueur)
            {

                for(k1=0;k1<nbNoeuds1*nbNoeuds1;k1++) 
                {
                    valeur3 = *(mxGetPr(MDist1)+k1);
                    if( valeur3 == longueur)
                    {
                        temp = comparerChemins(mxGetCell(chemin1,k1),mxGetCell(chemin2,j),  noyauNoeud, noyauArc,nbNoeuds1,nbArcs1);
                        if( maxi2 < temp)
                        {
                            maxi2=temp;
                        }
                    }
                }
                somme=somme+maxi2;
            }
        maxi2=-100000000;
        }
    }
    K = K+somme;
    }
    K=0.5*K;
    plhs[0]=mxCreateDoubleMatrix(1,1,mxREAL);
  mxGetPr(plhs[0])[0] = K;

  free(sigmaNoeud);
  free(sigmaArc);
  free(noyauNoeud);
  free(noyauArc);
  free(histoChemins1);
  free(histoChemins2);
}

 /********************************************************/
/*       	 fonction de comparaison de chemins              	    */
/********************************************************/
double comparerChemins(mxArray *chemin1, mxArray *chemin2 , double *noyauNoeud, double *noyauArc, int nbNoeud, int nbArc)
{
	int i,j,indice1,indice2;
	double produit;
	produit = 1.0;
    
    for(i=0;i<=mxGetN(chemin1);i=i+2)
    {
        indice1 = *(mxGetPr(chemin1)+i)-1;
        indice2 = *(mxGetPr(chemin2)+i)-1;
        produit = produit*noyauNoeud[indice2*nbNoeud + indice1];
    }

    for(i=1;i<mxGetN(chemin1);i=i+2)
    {
        indice1 = *(mxGetPr(chemin1)+i)-1;
        indice2 = *(mxGetPr(chemin2)+i)-1;
        produit = produit*noyauArc[indice2*nbArc + indice1];
    }

	return produit;
}


 
 
/********************************************************/
/*       fonction de calcul de noyau des noeuds         */
/********************************************************/
void calculerNoyauNoeuds(mxArray *etiquetteNoeud1 , mxArray *etiquetteNoeud2, double *sigmaNoeud, double *noyauNoeud)
{
    int nbNoeuds1, nbNoeuds2, nbEtiquettesNoeuds;
    int i1,i2,e;
    double somme;

    nbNoeuds1 = mxGetM(etiquetteNoeud1);
    nbNoeuds2 = mxGetM(etiquetteNoeud2);
    nbEtiquettesNoeuds = mxGetN(etiquetteNoeud1);

    for(i1=0;i1<nbNoeuds1;i1++)
    {
        for(i2=0;i2<nbNoeuds2;i2++)
        {
            somme = 0.0;
            for(e=0;e<nbEtiquettesNoeuds;e++)
            {
                somme = somme + pow( (*(mxGetPr(etiquetteNoeud1)+(e*nbNoeuds1+i1)) - *(mxGetPr(etiquetteNoeud2)+(e*nbNoeuds2+i2)))/sigmaNoeud[e],2);
            }
            /* noyauNoeud[i1*nbNoeuds1+i2] = exp(-somme); */
            noyauNoeud[i2*nbNoeuds1+i1] = exp(-somme);
        }
    }

}


/********************************************************/
/*       	 fonction de calcul de noyau des arcs                	    */
/********************************************************/
void calculerNoyauArcs(mxArray *etiquetteArc1 , mxArray *etiquetteArc2, double *sigmaArc, double *noyauArc)
{
    int nbArcs1, nbArcs2, nbEtiquettesArcs;
    int i1,i2,e;
    double somme;

    nbArcs1 = mxGetM(etiquetteArc1);
    nbArcs2 = mxGetM(etiquetteArc2);
    nbEtiquettesArcs = mxGetN(etiquetteArc1);
    
    for(i1=0;i1<nbArcs1;i1++)
    {
        for(i2=0;i2<nbArcs2;i2++)
        {
            somme = 0.0;
            for(e=0;e<nbEtiquettesArcs;e++)
            {
                        somme = somme + pow( (*(mxGetPr(etiquetteArc1)+(e*nbArcs1+i1)) - *(mxGetPr(etiquetteArc2)+(e*nbArcs2+i2)))/sigmaArc[e],2);
            }
            /* *(noyauArc +(i1*nbArcs1+i2)*sizeof(double)) = exp(-somme); */
            noyauArc[i2*nbArcs1+i1] = exp(-somme);	
        }
    }

}

/********************************************************/
/*       	 fonction de calcul du max Kc                	    */
/********************************************************/