function [MDist,Chemin] = calculerCheminsGraphe(MAdj,Distance)
%
% function [MDist, Chemin] = calculerCheminsGraphe(MAdj,Distance)
% fonction permettant de determiner les chemins les plus courts dans un graphe selon l'algo dijkstra
% MAdj : matrice d'adjacence du graphe
% Distance : distance entre chaque noeud
%
%%%%%%%%%%%%%%%%%%%
% 03/2006 F-Suard %
%%%%%%%%%%%%%%%%%%%

nbpts = length(MAdj);

%distance les plus courtes
MDist = 1-eye(nbpts);
MDist(MDist == 1) = Inf;

%recherche du voisinage
Voisinage = cell(nbpts,1);
for i=1:nbpts
    ind = find(MAdj(i,:) > 0);
    Voisinage{i} = ind;
end

Chemin = cell(nbpts);
%initialisation
for i=1:nbpts
    Chemin{i,i} = i;
    ind = Voisinage{i};
    MDist(i,ind) = min(MDist(i,ind),Distance(i,ind));
    for j=1:length(ind)
        Chemin{i,ind(j)} = [i, ind(j)];
    end
end


% on remplit MDist
% tant qu'il reste un chemin inf dans MDist
while sum(sum(isinf(MDist))) >  0
    for i=1:nbpts
        ind = find(MDist(i,:) ~= Inf &  MDist(i,:) > 0);
        for j=1:length(ind)
            indj = find(MDist(ind(j),:)~=Inf & MDist(ind(j),:) > 0);
            [vectmin indmin] = min([MDist(i,indj) ; MDist(i,ind(j)) + MDist(ind(j),indj)],[],1);
            indok = find(indmin ~= 1);
            for k=indok
                MDist(i,indj(k)) = min(MDist(i,indj(k)) , MDist(i,ind(j)) + MDist(ind(j),indj(k)));
                Chemin{i,indj(k)} = [Chemin{i,ind(j)}  Chemin{ind(j),indj(k)}(2:end)]; %MAdj(ind(j),indj(k))
            end
        end
    end
end


longueurChemin = zeros(nbpts);
for i=1:nbpts
        for j=1:nbpts
        longueurChemin(i,j) = size(Chemin{i,j},2);
    end
end



for i=1:size(Chemin,1)
    for j=1:size(Chemin,2)
        if length(Chemin{i,j}) > 1
            temp = Chemin{i,j};
            Chemin{i,j} = temp(1);
            for k=1:length(temp)-1
                Chemin{i,j} = [Chemin{i,j}, MAdj(temp(k),temp(k+1)),temp(k+1)];
            end
        end
    end
end