classdef simpleModelTSVM < simpleModel
    % SIMPLEMODELTSVM This class inherits from simpleModel. It contains
    % particular methods to solve binary-T-SVM.
    % This class support T-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 20/10/2011
    % Last Modified 20/10/2011 by G.L.
    
    properties
        e; % signe des multiplicateurs
        new; % indice of the latest SV
        cas;
    end
    
    methods
        % contructor
        function SM = simpleModelTSVM(SD,C,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelCSVM object
            
            assert(nargin==2 ,'simpleModelTSVM:wrongInput','Requires 2');
            if nargin<2
                error('not enough input');
            else
                SM.SD = SD;
                SM.C = C;
                SM.cas = 1;
                SM.b = 0;
                %                SM.Slacks = zeros(SD.n,1);
                %                 if nargin==5
                %                     SM.Il = Il;
                %                     SM.Iu = Iu;
                %                     SM.Iw = Iw;
                %                     SM.oldalpha = zeros(length(Iw),1);
                %                 else
                %                     SM.Iw = [find(SD.trainlab==-1,1),find(SD.trainlab==1,1)];
                %                     SM.Iu = [];
                %                     SM.Il = 1:SD.n;
                %                     SM.Il = setdiff(SM.Il,SM.Iw);
                %                     SM.oldalpha = zeros(2,1);
                %                 end
                SM.labs = sort(unique(SD.trainlab));
                SM.e = ones(size(SD.trainlab));
                %                 if length(C)==length(SD.trainlab)
                %                     SM.Slacks = C;
                %                 elseif length(C)==length(SM.labs)
                %                     for i=1:length(SM.labs)
                %                         SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                %                     end
                %                 else
                SM.Slacks = C(1)*ones(size(SD.trainlab));
                
                %                end
                SM.initialize;
                SM.activated = 0;
                SM.converged = 0;
                SM.cout = [];
            end
        end
        
        
        function set.e(obj,val)
            obj.e = val;
        end
        
        function val = get.e(obj)
            val = obj.e;
        end
        
        function set.new(obj,val)
            obj.new = val;
        end
        
        function val = get.new(obj)
            val = obj.new;
        end
        
        function set.cas(obj,val)
            obj.cas = val;
        end
        
        function val = get.cas(obj)
            val = obj.cas;
        end
        
        function initialize(obj)
            ok = -1;
            while ok==-1
                m = randperm(obj.SD.n);
                Iw = sort([find(obj.SD.trainlab(m)==-1,2);find(obj.SD.trainlab(m)==1,2)])';
                Iw = m(Iw);
                Iwn = [];
                Iwp = Iw;
                pp = 1;
                pm = 1;
                while pp>0 && pm>0
                    pp = 1;
                    while pp>0
                        indd = obj.SD.getKernelValues(Iwp,Iwp);
                        [~,pp] = chol(obj.SD.kernelCache(Iwp,indd)+0.001*eye(length(Iwp)))
                        if pp==0
                            ok = 1;
                        else
                            Iwn = [Iwn,Iwp(pp)];
                            Iwp(pp) = [];
                        end
                    end
                end
                pm=1;
                while pm>0
                    indd = obj.SD.getKernelValues(Iwn,Iwn);
                    [~,pm] = chol(obj.SD.kernelCache(Iwn,indd)+0.001*eye(length(Iwn)))
                    if pm>0
                        Iwp = [Iwp,Iwn(pm)];
                        Iwn(pm) = [];
                    end
                end
            end
            
            obj.Slacks(Iwp) = abs(obj.Slacks(Iwp));%.*obj.SD.trainlab(Iwp);
            obj.Slacks(Iwn) = -abs(obj.Slacks(Iwn));%.*obj.SD.trainlab(Iwn);
            
            
            
            obj.Iw = sort([Iwp,Iwn]);
            %obj.Iw = [find(obj.SD.trainlab==-1,1),find(obj.SD.trainlab==1,1)];
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(length(obj.Iw),1);
            
            obj.solveLinearSystem();
            if sign(obj.alpha.*obj.Slacks(obj.Iw))==ones(size(obj.alpha))
                ok = 1;
            end
            %             [obj.alpha,obj.SD.trainlab(obj.Iw),obj.Slacks(obj.Iw)]
            %             keyboard
            % end
            
            
            disp('init fini');
            keyboard
        end
        
        function train(obj)
            iter = 0;
            while obj.converged==0 && iter<1000*length(obj.SD.n)
                iter = iter+1;
                
                obj.solveLinearSystem;
                %                     obj.b
                %                     obj.cas = -obj.cas;
                %                     if (~isempty(intersect(obj.Iw,find(sign(obj.Slacks)==obj.cas))))
                %                         obj.new = 0;
                %                         obj.solveLinearSystem;
                %                         obj.b
                %                         obj.cas = -obj.cas;
                %                         obj.solveLinearSystem;
                %                         obj.b
                %                         keyboard
                %                     end
                
                
                obj.print2D();
                disp([obj.Iw',obj.alpha,sign(obj.Slacks(obj.Iw))]);
                disp(obj.new)
                obj.new = 0;
                %disp(evaluateSolution(obj,obj.Iw));
                keyboard
                obj.activateBoxConstraints();
                if (obj.activated==0)
                    indd = obj.SD.getKernelValues([obj.Iw,obj.Iu],[obj.Iw,obj.Iu]);
                    obj.cout(1,end+1) = -0.5*([obj.alpha;obj.Slacks(obj.Iu)].*obj.e([obj.Iw,obj.Iu]))'...
                        *obj.SD.kernelCache([obj.Iw,obj.Iu],indd)...
                        *([obj.alpha;obj.Slacks(obj.Iu)].*obj.e([obj.Iw,obj.Iu]))...
                        +sum(([obj.alpha;obj.Slacks(obj.Iu)].*obj.e([obj.Iw,obj.Iu])))...
                        -obj.b*obj.SD.trainlab([obj.Iw,obj.Iu])'*([obj.alpha;obj.Slacks(obj.Iu)].*obj.e([obj.Iw,obj.Iu]));
                    obj.cout([obj.Iw,obj.Iu]+1,end) = ([obj.alpha;obj.Slacks(obj.Iu)]'*obj.SD.kernelCache([obj.Iw,obj.Iu],indd)-1)';
                    %obj.cout = [obj.cout, [-0.5*obj.alpha'*obj.SD.kernelCache(obj.Iw,indd)*obj.alpha+sum(obj.alpha); (-obj.alpha'*obj.SD.kernelCache(obj.Iw,indd)+1)']];
                    
                    obj.relaxBoxConstraints();
                end
            end
            if obj.converged==0
                disp('No convergence');
                obj.solveLinearSystem();
                obj.converged = 1;
            end
            obj.finish();
            %obj.print2D();
            
        end
        
        function relaxBoxConstraints(obj)
            xt = zeros(obj.SD.n,1);
            xt(obj.Iw) = obj.alpha;
            xt(obj.Iu) = obj.Slacks(obj.Iu);
            S = 500;
            Found=0;
            nbLoop = ceil(length(obj.Il)/S);
            z = 0;
            if not(isempty(obj.Il))
                while (z<nbLoop && Found==0)
                    z = z+1;
                    nul = obj.evaluateSolution(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end)))-1;
                    [mm mpos]=min(nul);
                    if (mm<-obj.kktgap)
                        Found=1;
                        mpos = mpos+(obj.iloop-1)*S;
                    else
                        mm = inf; mpos = [];
                    end
                    %obj.iloop= obj.iloop+1;
                    obj.iloop = mod(obj.iloop,nbLoop)+1;
                    %keyboard
                end
            else
                mm = inf; mpos = [];
            end
            if not(isempty(obj.Iu))
                nuu = obj.evaluateSolution(obj.Iu)-1;
                [mmS mposS]=min(-nuu);
            else
                mmS = inf; mposS = [];
            end
            
            if (~isempty(mmS) && mmS < -obj.kktgap) || (~isempty(mm) && mm < -obj.kktgap)
                if isempty(mmS) || mm < mmS
                    obj.Iw = sort([obj.Iw , obj.Il(mpos)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.new = obj.Il(mpos);
                    obj.Il(mpos) = [];
                else
                    obj.Iw = sort([obj.Iw , obj.Iu(mposS)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.new = obj.Iu(mposS);
                    obj.Iu(mposS)=[];
                end
            else
                obj.converged = 1;
            end
        end
        
        function activateBoxConstraints(obj) % on ne projette pas dans le meme sens selon le sign de alpha...
            intersect(obj.Iw,find(obj.Slacks>0))
            intersect(obj.Iw,find(obj.Slacks<0))
            if (~isempty(find(sign(obj.alpha)==-sign(obj.Slacks(obj.Iw)), 1)) || ...
                ~isempty(find(abs(obj.alpha)>abs(obj.Slacks(obj.Iw)), 1))) && (...
                length(intersect(obj.Iw,find(obj.Slacks>0)))>3 || ...
                length(intersect(obj.Iw,find(obj.Slacks<0)))>3)
                % if (~isempty(find(abs(obj.alpha)>abs(obj.Slacks(obj.Iw)), 1))) && length(obj.alpha)>5
                dir = obj.alpha - obj.oldalpha;
                indad = find(sign(obj.alpha)==-sign(obj.Slacks(obj.Iw)));
                %indad = [];
                indsup = find(abs(obj.alpha) > abs(obj.Slacks(obj.Iw)));
                indsup = setdiff(indsup,indad);
                [tI indmin] = min(abs(-obj.oldalpha(indad))./dir(indad));
                [tS indS] = min(abs((obj.Slacks(obj.Iw(indsup))-obj.oldalpha(indsup)))./dir(indsup));
                if isempty(tI) , tI = tS + 1; end;
                if isempty(tS) , tS = tI + 1; end;
                t = min(tI,tS);
                obj.oldalpha = obj.oldalpha + t*dir;
                if t==tI
                    obj.Il = sort([obj.Il , obj.Iw(indad(indmin))]);
                    obj.Iw(indad(indmin)) = [];
                    obj.oldalpha(indad(indmin)) = [];
                    obj.alpha(indad(indmin)) = [];
                else
                    obj.Iu = sort([obj.Iu , obj.Iw(indsup(indS))]);
                    obj.Iw(indsup(indS))= [];
                    obj.oldalpha(indsup(indS))= [];
                    obj.alpha(indsup(indS)) = [];
                end
                obj.activated = 1;
            else
                obj.activated = 0;
            end
            
        end
        
        
        function solveLinearSystem(obj)
            % test sur la positivite du groupe +
            if obj.new>0
                ip = intersect(obj.Iw,find(obj.Slacks>0));
                indd = obj.SD.getKernelValues(ip,obj.Iu);
                [~,p] = chol(obj.SD.kernelCache(ip,indd)+0.1*eye(length(ip)));
                if p>0
                    ip = intersect(obj.Iw,find(obj.Slacks<0));
                    indd = obj.SD.getKernelValues(ip,obj.Iu);
                    [~,p] = chol(obj.SD.kernelCache(ip,indd)+0.001*eye(length(ip)));
                    p
                    %if p==0
                        obj.Slacks(obj.new) = -(abs(obj.Slacks(obj.new)));
                        %obj.cas = -1;
                    %end
                else
                    obj.Slacks(obj.new) = abs(obj.Slacks(obj.new));
                    %  obj.cas = 1;
                end
                
            end
            
            if obj.cas==1
                points = obj.Iw;%intersect(obj.Iw,find(obj.Slacks>0));
                autres = [];%intersect(obj.Iw,find(obj.Slacks<0));
                
            elseif obj.cas==-1
                points = intersect(obj.Iw,find(obj.Slacks<0));
                autres = intersect(obj.Iw,find(obj.Slacks>0));
                
            end
            
            [~,ipoints] = intersect(obj.Iw,points);
            ipoints = ipoints';
            
            [~,iautres] = intersect(obj.Iw,autres);
            iautres = iautres';
            
            
            ce = ones(1,length(points));
            be = 0;
            
            indd = obj.SD.getKernelValues(points,[obj.Iu;autres]);
            if ~isempty(obj.Iu)
                ce = ce - (obj.Slacks(obj.Iu))'*obj.SD.kernelCache(obj.Iu,indd);
                be = be - (obj.Slacks(obj.Iu))'*obj.SD.trainlab(obj.Iu);
            end
            if ~isempty(autres)
                ce = ce - (obj.oldalpha(iautres))'*obj.SD.kernelCache(autres,indd);
                be = be - (obj.oldalpha(iautres))'*obj.SD.trainlab(autres);
            end
            obj.alpha = obj.oldalpha;
            if obj.optimType==1
                [L,U] = lu(obj.SD.kernelCache(points,indd)+0.001*eye(length(points)));
                M = (obj.SD.trainlab(points))'*(U\(L\(obj.SD.trainlab(points))));
                obj.b = M\((obj.SD.trainlab(points))'*(U\(L\ce'))-be);
                obj.alpha(ipoints) = (U\(L\(ce'-obj.SD.trainlab(points)*obj.b)));
                obj.alpha = reshape(obj.alpha,length(obj.alpha),1);
            else
                error('not implemented yet');
            end
            
            
            %             if obj.new > 0
            %                 i = obj.Iw==obj.new;
            %                 if (obj.alpha(i)*obj.SD.trainlab(obj.new)<0)
            %                     obj.Slacks(obj.new) = -abs(obj.Slacks(obj.new));
            %                 end
            %             end
            
        end
        
        
        function reinit(obj)
            obj.Iw = [find(obj.SD.trainlab==-1,1),find(obj.SD.trainlab==1,1)];
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelTSVM:print2D','must be 2D data');
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
            f = kernelEval(obj,xtest);
            if obj.converged==0
                ypred = ([obj.alpha;obj.Slacks(obj.Iu)].*obj.SD.trainlab([obj.Iw,obj.Iu]))'*f+obj.b;
            else
                ypred = obj.alpha'*f+obj.b;
            end
            ypredmat=reshape(ypred,na,nb);
            
            hold off;
            plot(obj.SD.trainvec(1,obj.SD.trainlab==1),obj.SD.trainvec(2,obj.SD.trainlab==1),'k+');
            hold on;
            plot(obj.SD.trainvec(1,obj.SD.trainlab==-1),obj.SD.trainvec(2,obj.SD.trainlab==-1),'ko');
            plot(obj.SD.trainvec(1,[obj.Iw]),obj.SD.trainvec(2,[obj.Iw]),'om','MarkerSize',6,'LineWidth',2);
            plot(obj.SD.trainvec(1,[obj.Iu]),obj.SD.trainvec(2,[obj.Iu]),'oc','MarkerSize',6,'LineWidth',2);
            plot(obj.SV(1,obj.alpha>0),obj.SV(2,obj.alpha>0),'ob');
            plot(obj.SV(1,obj.alpha<0),obj.SV(2,obj.alpha<0),'or');
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ 0 0],'g');
            set(h,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ -1 1],'y.');
            set(h,'LineWidth',1);
            drawnow;
            
        end
        
        
        
        function print2DMulti(obj,SMt)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelTSVM:print2D','must be 2D data');
            assert(nargin==2);
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            if length(SMt)==length(unique(obj.SD.trainlab));
                ypred = obj.test1vsAll(SMt,xtest);
            else
                ypred = obj.test1vs1(SMt,xtest);
            end
            ypredmat=reshape(ypred,na,nb);
            %labs = unique(obj.SD.trainlab);
            hold on;
            [c,h] = contour(xtesta1,xtesta2,ypredmat);
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        
        function finish(obj)
            if ~isempty(obj.SD.trainvec)
                obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
                
            else
                %disp('no training data can be stored for later use :
                %pre-computed matrix');
            end
            obj.alpha = [obj.alpha;obj.Slacks(obj.Iu)].*obj.SD.trainlab([obj.Iw,obj.Iu]);
        end
        
        
        
        
        
    end
    
end
